@@ -36,6 +36,22 @@ static ap_filter_rec_t *cache_out_subreq_filter_handle;
 static ap_filter_rec_t *cache_remove_url_filter_handle;
 static ap_filter_rec_t *cache_invalidate_filter_handle;
 
+/**
+ * Entity headers' names
+ */
+static const char *MOD_CACHE_ENTITY_HEADERS[] = {
+    "Allow",
+    "Content-Encoding",
+    "Content-Language",
+    "Content-Length",
+    "Content-Location",
+    "Content-MD5",
+    "Content-Range",
+    "Content-Type",
+    "Last-Modified",
+    NULL
+};
+
 /*
  * CACHE handler
  * -------------
@@ -218,6 +234,11 @@ static int cache_quick_handler(request_rec *r, int lookup)
                     ap_log_rerror(APLOG_MARK, APLOG_DEBUG, rv,
                             r, APLOGNO(00752) "Cache locked for url, not caching "
                             "response: %s", r->uri);
+                    /* cache_select() may have added conditional headers */
+                    if (cache->stale_headers) {
+                        r->headers_in = cache->stale_headers;
+                    }
+
                 }
             }
             else {
@@ -620,7 +641,6 @@ static int cache_handler(request_rec *r)
 static apr_status_t cache_out_filter(ap_filter_t *f, apr_bucket_brigade *in)
 {
     request_rec *r = f->r;
-    apr_bucket *e;
     cache_request_rec *cache = (cache_request_rec *)f->ctx;
 
     if (!cache) {
@@ -636,10 +656,8 @@ static apr_status_t cache_out_filter(ap_filter_t *f, apr_bucket_brigade *in)
             "cache: running CACHE_OUT filter");
 
     /* clean out any previous response up to EOS, if any */
-    for (e = APR_BRIGADE_FIRST(in);
-         e != APR_BRIGADE_SENTINEL(in);
-         e = APR_BUCKET_NEXT(e))
-    {
+    while (!APR_BRIGADE_EMPTY(in)) {
+        apr_bucket *e = APR_BRIGADE_FIRST(in);
         if (APR_BUCKET_IS_EOS(e)) {
             apr_bucket_brigade *bb = apr_brigade_create(r->pool,
                     r->connection->bucket_alloc);
@@ -697,7 +715,7 @@ static int cache_save_store(ap_filter_t *f, apr_bucket_brigade *in,
         rv = cache->provider->store_body(cache->handle, f->r, in, cache->out);
         if (rv != APR_SUCCESS) {
             ap_log_rerror(APLOG_MARK, APLOG_DEBUG, rv, f->r, APLOGNO(00765)
-                    "cache: Cache provider's store_body failed!");
+                    "cache: Cache provider's store_body failed for URI %s", f->r->uri);
             ap_remove_output_filter(f);
 
             /* give someone else the chance to cache the file */
@@ -756,17 +774,16 @@ static int cache_save_store(ap_filter_t *f, apr_bucket_brigade *in,
 /**
  * Sanity check for 304 Not Modified responses, as per RFC2616 Section 10.3.5.
  */
-static const char *cache_header_cmp(apr_pool_t *pool, apr_table_t *left,
+static int cache_header_cmp(apr_pool_t *pool, apr_table_t *left,
         apr_table_t *right, const char *key)
 {
     const char *h1, *h2;
 
     if ((h1 = cache_table_getm(pool, left, key))
             && (h2 = cache_table_getm(pool, right, key)) && (strcmp(h1, h2))) {
-        return apr_pstrcat(pool, "contradiction: 304 Not Modified, but ", key,
-                " modified", NULL);
+        return 1;
     }
-    return NULL;
+    return 0;
 }
 
 /*
@@ -802,7 +819,7 @@ static apr_status_t cache_save_filter(ap_filter_t *f, apr_bucket_brigade *in)
     apr_time_t exp, date, lastmod, now;
     apr_off_t size = -1;
     cache_info *info = NULL;
-    const char *reason;
+    const char *reason, **eh;
     apr_pool_t *p;
     apr_bucket *e;
     apr_table_t *headers;
@@ -1104,37 +1121,29 @@ static apr_status_t cache_save_filter(ap_filter_t *f, apr_bucket_brigade *in)
         reason = "updated entity is older than cached entity";
 
         /* while this response is not cacheable, the previous response still is */
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(00770)
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(02474)
                 "cache: Removing CACHE_REMOVE_URL filter.");
         ap_remove_output_filter(cache->remove_url_filter);
     }
     else if (r->status == HTTP_NOT_MODIFIED && cache->stale_handle) {
         apr_table_t *left = cache->stale_handle->resp_hdrs;
         apr_table_t *right = r->headers_out;
+        const char *ehs = NULL;
 
         /* and lastly, contradiction checks for revalidated responses
          * as per RFC2616 Section 10.3.5
          */
-        if (((reason = cache_header_cmp(r->pool, left, right, "Allow")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Content-Encoding")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Content-Language")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Content-Length")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Content-Location")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Content-MD5")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Content-Range")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Content-Type")))
-                || ((reason = cache_header_cmp(r->pool, left, right, "Expires")))
-                || ((reason = cache_header_cmp(r->pool, left, right, "ETag")))
-                || ((reason = cache_header_cmp(r->pool, left, right,
-                        "Last-Modified")))) {
-            /* contradiction: 304 Not Modified, but entity header modified */
+        if (cache_header_cmp(r->pool, left, right, "ETag")) {
+            ehs = "ETag";
+        }
+        for (eh = MOD_CACHE_ENTITY_HEADERS; *eh; ++eh) {
+            if (cache_header_cmp(r->pool, left, right, *eh)) {
+                ehs = (ehs) ? apr_pstrcat(r->pool, ehs, ", ", *eh, NULL) : *eh;
+            }
+        }
+        if (ehs) {
+            reason = apr_pstrcat(r->pool, "contradiction: 304 Not Modified; "
+                                 "but ", ehs, " modified", NULL);
         }
     }
 
@@ -1149,14 +1158,9 @@ static apr_status_t cache_save_filter(ap_filter_t *f, apr_bucket_brigade *in)
      * inconsistencies between cached entity-bodies and updated headers.
      */
     if (r->status == HTTP_NOT_MODIFIED) {
-        apr_table_unset(r->headers_out, "Allow");
-        apr_table_unset(r->headers_out, "Content-Encoding");
-        apr_table_unset(r->headers_out, "Content-Language");
-        apr_table_unset(r->headers_out, "Content-Length");
-        apr_table_unset(r->headers_out, "Content-MD5");
-        apr_table_unset(r->headers_out, "Content-Range");
-        apr_table_unset(r->headers_out, "Content-Type");
-        apr_table_unset(r->headers_out, "Last-Modified");
+        for (eh = MOD_CACHE_ENTITY_HEADERS; *eh; ++eh) {
+            apr_table_unset(r->headers_out, *eh);
+        }
     }
 
     /* Hold the phone. Some servers might allow us to cache a 2xx, but
@@ -1171,8 +1175,10 @@ static apr_status_t cache_save_filter(ap_filter_t *f, apr_bucket_brigade *in)
      */
     if (reason && r->status == HTTP_NOT_MODIFIED && cache->stale_handle) {
 
-        ap_log_rerror(
-                APLOG_MARK, APLOG_INFO, 0, r, APLOGNO() "cache: %s responded with an uncacheable 304, retrying the request. Reason: %s", r->unparsed_uri, reason);
+        ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, APLOGNO(02473) 
+                "cache: %s responded with an uncacheable 304, " 
+                "retrying the request. Reason: %s", 
+                r->unparsed_uri, reason);
 
         /* we've got a cache conditional miss! tell anyone who cares */
         cache_run_cache_status(cache->handle, r, r->headers_out, AP_CACHE_MISS,
@@ -1197,7 +1203,9 @@ static apr_status_t cache_save_filter(ap_filter_t *f, apr_bucket_brigade *in)
         apr_table_unset(r->headers_in, "If-Range");
         apr_table_unset(r->headers_in, "If-Unmodified-Since");
 
-        ap_internal_redirect(r->uri, r);
+        /* Currently HTTP_NOT_MODIFIED, and after the redirect, handlers won't think to set status to HTTP_OK */
+        r->status = HTTP_OK; 
+        ap_internal_redirect(r->unparsed_uri, r);
 
         return APR_SUCCESS;
     }
@@ -1438,10 +1446,12 @@ static apr_status_t cache_save_filter(ap_filter_t *f, apr_bucket_brigade *in)
          * the cached headers.
          *
          * However, before doing that, we need to first merge in
-         * err_headers_out and we also need to strip any hop-by-hop
-         * headers that might have snuck in.
+         * err_headers_out (note that store_headers() below already selects
+         * the cacheable only headers using ap_cache_cacheable_headers_out(),
+         * here we want to keep the original headers in r->headers_out and
+         * forward all of them to the client, including non-cacheable ones).
          */
-        r->headers_out = ap_cache_cacheable_headers_out(r);
+        r->headers_out = cache_merge_headers_out(r);
 
         /* Merge in our cached headers.  However, keep any updated values. */
         /* take output, overlay on top of cached */
@@ -1491,6 +1501,13 @@ static apr_status_t cache_save_filter(ap_filter_t *f, apr_bucket_brigade *in)
         if (status != OK) {
             r->status = status;
 
+            /* Strip the entity headers merged from the cached headers before
+             * updating the entry (see cache_accept_headers() above).
+             */
+            for (eh = MOD_CACHE_ENTITY_HEADERS; *eh; ++eh) {
+                apr_table_unset(r->headers_out, *eh);
+            }
+
             bkt = apr_bucket_flush_create(bb->bucket_alloc);
             APR_BRIGADE_INSERT_TAIL(bb, bkt);
         }
@@ -2070,8 +2087,7 @@ static const char *set_cache_quick_handler(cmd_parms *parms, void *dummy,
     cache_server_conf *conf;
 
     conf =
-        (cache_server_conf *)ap_get_module_config(parms->server->module_config
-,
+        (cache_server_conf *)ap_get_module_config(parms->server->module_config,
                                                   &cache_module);
     conf->quick = flag;
     conf->quick_set = 1;
