@@ -22,6 +22,7 @@
 #include "http_config.h"
 #include "http_log.h"
 #include "http_core.h"
+#include "http_protocol.h"
 #include "ap_provider.h"
 #include "ap_socache.h"
 #include "util_filter.h"
@@ -30,6 +31,7 @@
 #include "util_mutex.h"
 
 #include "mod_cache.h"
+#include "mod_status.h"
 
 #include "cache_socache_common.h"
 
@@ -168,8 +170,8 @@ static apr_status_t store_array(apr_array_header_t *arr, unsigned char *buffer,
     elts = (const char **) arr->elts;
 
     for (i = 0; i < arr->nelts; i++) {
-        len = strlen(elts[i]);
-        if (len + 3 >= buffer_len - *slider) {
+        apr_size_t e_len = strlen(elts[i]);
+        if (e_len + 3 >= buffer_len - *slider) {
             return APR_EOF;
         }
         len = apr_snprintf(buffer ? (char *) buffer + *slider : NULL,
@@ -919,7 +921,7 @@ static apr_status_t store_headers(cache_handle_t *h, request_rec *r,
     if (sobj->headers_in) {
         if (APR_SUCCESS != store_table(sobj->headers_in, sobj->buffer,
                 sobj->buffer_len, &slider)) {
-            ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r, APLOGNO(02376)
+            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r, APLOGNO(02376)
                     "in-headers didn't fit in buffer %s",
                     sobj->key);
             apr_pool_destroy(sobj->pool);
@@ -1239,10 +1241,11 @@ static void *create_config(apr_pool_t *p, server_rec *s)
 
 static void *merge_config(apr_pool_t *p, void *basev, void *overridesv)
 {
-    cache_socache_conf *ps = apr_pcalloc(p, sizeof(cache_socache_conf));
+    cache_socache_conf *ps;
     cache_socache_conf *base = (cache_socache_conf *) basev;
     cache_socache_conf *overrides = (cache_socache_conf *) overridesv;
 
+    /* socache server config only has one field */
     ps = overrides ? overrides : base;
 
     return ps;
@@ -1374,6 +1377,56 @@ static apr_status_t destroy_cache(void *data)
     return APR_SUCCESS;
 }
 
+static int socache_status_hook(request_rec *r, int flags)
+{
+    apr_status_t status = APR_SUCCESS;
+    cache_socache_conf *conf = ap_get_module_config(r->server->module_config,
+                                                    &cache_socache_module);
+    if (!conf->provider || !conf->provider->socache_provider ||
+        !conf->provider->socache_instance) {
+        return DECLINED;
+    }
+
+    ap_rputs("<hr>\n"
+             "<table cellspacing=0 cellpadding=0>\n"
+             "<tr><td bgcolor=\"#000000\">\n"
+             "<b><font color=\"#ffffff\" face=\"Arial,Helvetica\">"
+             "mod_cache_socache Status:</font></b>\n"
+             "</td></tr>\n"
+             "<tr><td bgcolor=\"#ffffff\">\n", r);
+
+    if (socache_mutex) {
+        status = apr_global_mutex_lock(socache_mutex);
+        if (status != APR_SUCCESS) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r, APLOGNO(02816)
+                    "could not acquire lock for cache status");
+        }
+    }
+
+    if (status != APR_SUCCESS) {
+        ap_rputs("No cache status data available\n", r);
+    } else {
+        conf->provider->socache_provider->status(conf->provider->socache_instance,
+                                                 r, flags);
+    }
+
+    if (socache_mutex && status == APR_SUCCESS) {
+        status = apr_global_mutex_unlock(socache_mutex);
+        if (status != APR_SUCCESS) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r, APLOGNO(02817)
+                    "could not release lock for cache status");
+        }
+    }
+
+    ap_rputs("</td></tr>\n</table>\n", r);
+    return OK;
+}
+
+static void socache_status_register(apr_pool_t *p)
+{
+    APR_OPTIONAL_HOOK(ap, status_hook, socache_status_hook, NULL, NULL, APR_HOOK_MIDDLE);
+}
+
 static int socache_precfg(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptmp)
 {
     apr_status_t rv = ap_mutex_register(pconf, cache_socache_id, NULL,
@@ -1383,6 +1436,10 @@ static int socache_precfg(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptmp)
         "failed to register %s mutex", cache_socache_id);
         return 500; /* An HTTP status would be a misnomer! */
     }
+
+    /* Register to handle mod_status status page generation */
+    socache_status_register(pconf);
+
     return OK;
 }
 
@@ -1393,7 +1450,7 @@ static int socache_post_config(apr_pool_t *pconf, apr_pool_t *plog,
     apr_status_t rv;
     const char *errmsg;
     static struct ap_socache_hints socache_hints =
-    { 64, 32, 60000000 };
+    { 64, 2048, 60000000 };
 
     for (s = base_server; s; s = s->next) {
         cache_socache_conf *conf =
