@@ -15,7 +15,7 @@
  */
 
 /*
- * Apache example module.  Provide demonstrations of how modules do things.
+ * Apache example_hooks module.  Provide demonstrations of how modules do things.
  * It is not meant to be used in a production server.  Since it participates
  * in all of the processing phases, it could conceivable interfere with
  * the proper operation of other modules -- particularly the ones related
@@ -26,7 +26,7 @@
  * prefixed with 'x_' instead of 'example_'.
  *
  * To use mod_example_hooks, configure the Apache build with
- * --enable-example and compile.  Set up a <Location> block in your
+ * --enable-example-hooks and compile.  Set up a <Location> block in your
  * configuration file like so:
  *
  * <Location /example>
@@ -328,11 +328,11 @@ static x_cfg *our_cconfig(const conn_rec *c)
 static void example_log_each(apr_pool_t *p, server_rec *s, const char *note)
 {
     if (s != NULL) {
-        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_example: %s", note);
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_example_hooks: %s", note);
     } else {
         apr_file_t *out = NULL;
         apr_file_open_stderr(&out, p);
-        apr_file_printf(out, "mod_example traced in non-loggable "
+        apr_file_printf(out, "mod_example_hooks traced in non-loggable "
                         "context: %s\n", note);
     }
 }
@@ -703,7 +703,7 @@ static void *x_merge_server_config(apr_pool_t *p, void *server1_conf,
  * declaration near the bottom of this file.)  Note that these may be       *
  * called for situations that don't relate primarily to our function - in   *
  * other words, the fixup handler shouldn't assume that the request has     *
- * to do with "example" stuff.                                              *
+ * to do with "example_hooks" stuff.                                        *
  *                                                                          *
  * With the exception of the content handler, all of our routines will be   *
  * called for each request, unless an earlier handler from another module   *
@@ -1019,7 +1019,7 @@ static int x_handler(request_rec *r)
             ap_get_server_banner());
     ap_rputs("  <BR>\n", r);
     ap_rprintf(r, "  Server built: \"%s\"\n", ap_get_server_built());
-    ap_rputs("  </P>\n", r);;
+    ap_rputs("  </P>\n", r);
     ap_rputs("  <P>\n", r);
     ap_rputs("  The format for the callback trace is:\n", r);
     ap_rputs("  </P>\n", r);
