@@ -142,8 +142,7 @@ static void fix_skipto(request_rec* r, xml2ctx* ctx)
                                                &bstart);
                     ap_assert(rv == APR_SUCCESS);
                     while (b = APR_BRIGADE_FIRST(ctx->bbsave), b != bstart) {
-                        APR_BUCKET_REMOVE(b);
-                        apr_bucket_destroy(b);
+                        apr_bucket_delete(b);
                     }
                     ctx->bytes -= (p-ctx->buf);
                     ctx->buf = p ;
@@ -228,8 +227,7 @@ static void sniff_encoding(request_rec* r, xml2ctx* ctx)
         /* cut out the <meta> we're invalidating */
         while (cutb != cute) {
             b = APR_BUCKET_NEXT(cutb);
-            APR_BUCKET_REMOVE(cutb);
-            apr_bucket_destroy(cutb);
+            apr_bucket_delete(cutb);
             cutb = b;
         }
         /* and leave a string */
@@ -370,6 +368,9 @@ static apr_status_t xml2enc_ffunc(ap_filter_t* f, apr_bucket_brigade* bb)
         /* nah, we only have one action here - call it inline */
         fix_skipto(f->r, ctx);
 
+        /* we might change the Content-Length, so let's force its re-calculation */
+        apr_table_unset(f->r->headers_out, "Content-Length");
+
         /* consume the data we just sniffed */
         /* we need to omit any <meta> we just invalidated */
         ctx->flags |= ENC_INITIALISED;
@@ -432,8 +433,7 @@ static apr_status_t xml2enc_ffunc(ap_filter_t* f, apr_bucket_brigade* bb)
                 /* remove the data we've just read */
                 rv = apr_brigade_partition(bb, bytes, &bstart);
                 while (b = APR_BRIGADE_FIRST(bb), b != bstart) {
-                    APR_BUCKET_REMOVE(b);
-                    apr_bucket_destroy(b);
+                    apr_bucket_delete(b);
                 }
                 ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, APLOGNO(01438)
                               "xml2enc: consuming %" APR_SIZE_T_FMT
