@@ -15,57 +15,14 @@
  */
 
 #include "mod_proxy.h"
-#include "fcgi_protocol.h"
+#include "util_fcgi.h"
 #include "util_script.h"
 
 module AP_MODULE_DECLARE_DATA proxy_fcgi_module;
 
-/*
- * The below 3 functions serve to map the FCGI structs
- * back and forth between an 8 byte array. We do this to avoid
- * any potential padding issues when we send or read these
- * structures.
- *
- * NOTE: These have specific internal knowledge of the
- *       layout of the fcgi_header and fcgi_begin_request_body
- *       structs!
- */
-static void fcgi_header_to_array(fcgi_header *h, unsigned char a[])
-{
-    a[FCGI_HDR_VERSION_OFFSET]        = h->version;
-    a[FCGI_HDR_TYPE_OFFSET]           = h->type;
-    a[FCGI_HDR_REQUEST_ID_B1_OFFSET]  = h->requestIdB1;
-    a[FCGI_HDR_REQUEST_ID_B0_OFFSET]  = h->requestIdB0;
-    a[FCGI_HDR_CONTENT_LEN_B1_OFFSET] = h->contentLengthB1;
-    a[FCGI_HDR_CONTENT_LEN_B0_OFFSET] = h->contentLengthB0;
-    a[FCGI_HDR_PADDING_LEN_OFFSET]    = h->paddingLength;
-    a[FCGI_HDR_RESERVED_OFFSET]       = h->reserved;
-}
-
-static void fcgi_header_from_array(fcgi_header *h, unsigned char a[])
-{
-    h->version         = a[FCGI_HDR_VERSION_OFFSET];
-    h->type            = a[FCGI_HDR_TYPE_OFFSET];
-    h->requestIdB1     = a[FCGI_HDR_REQUEST_ID_B1_OFFSET];
-    h->requestIdB0     = a[FCGI_HDR_REQUEST_ID_B0_OFFSET];
-    h->contentLengthB1 = a[FCGI_HDR_CONTENT_LEN_B1_OFFSET];
-    h->contentLengthB0 = a[FCGI_HDR_CONTENT_LEN_B0_OFFSET];
-    h->paddingLength   = a[FCGI_HDR_PADDING_LEN_OFFSET];
-    h->reserved        = a[FCGI_HDR_RESERVED_OFFSET];
-}
-
-static void fcgi_begin_request_body_to_array(fcgi_begin_request_body *h,
-                                             unsigned char a[])
-{
-    a[FCGI_BRB_ROLEB1_OFFSET]    = h->roleB1;
-    a[FCGI_BRB_ROLEB0_OFFSET]    = h->roleB0;
-    a[FCGI_BRB_FLAGS_OFFSET]     = h->flags;
-    a[FCGI_BRB_RESERVED0_OFFSET] = h->reserved[0];
-    a[FCGI_BRB_RESERVED1_OFFSET] = h->reserved[1];
-    a[FCGI_BRB_RESERVED2_OFFSET] = h->reserved[2];
-    a[FCGI_BRB_RESERVED3_OFFSET] = h->reserved[3];
-    a[FCGI_BRB_RESERVED4_OFFSET] = h->reserved[4];
-}
+typedef struct {
+    int need_dirwalk;
+} fcgi_req_config_t;
 
 /*
  * Canonicalise http-like URLs.
@@ -76,8 +33,11 @@ static void fcgi_begin_request_body_to_array(fcgi_begin_request_body *h,
 static int proxy_fcgi_canon(request_rec *r, char *url)
 {
     char *host, sport[7];
-    const char *err, *path;
-    apr_port_t port = 8000;
+    const char *err;
+    char *path;
+    apr_port_t port, def_port;
+    fcgi_req_config_t *rconf = NULL;
+    const char *pathinfo_type = NULL;
 
     if (strncasecmp(url, "fcgi:", 5) == 0) {
         url += 5;
@@ -86,9 +46,10 @@ static int proxy_fcgi_canon(request_rec *r, char *url)
         return DECLINED;
     }
 
+    port = def_port = ap_proxy_port_of_scheme("fcgi");
+
     ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r,
                  "canonicalising URL %s", url);
-
     err = ap_proxy_canon_netloc(r->pool, &url, NULL, NULL, &host, &port);
     if (err) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01059)
@@ -96,7 +57,10 @@ static int proxy_fcgi_canon(request_rec *r, char *url)
         return HTTP_BAD_REQUEST;
     }
 
-    apr_snprintf(sport, sizeof(sport), ":%d", port);
+    if (port != def_port)
+        apr_snprintf(sport, sizeof(sport), ":%d", port);
+    else
+        sport[0] = '\0';
 
     if (ap_strchr_c(host, ':')) {
         /* if literal IPv6 address */
@@ -119,67 +83,67 @@ static int proxy_fcgi_canon(request_rec *r, char *url)
     ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01060)
                   "set r->filename to %s", r->filename);
 
-    if (apr_table_get(r->subprocess_env, "proxy-fcgi-pathinfo")) {
-        r->path_info = apr_pstrcat(r->pool, "/", path, NULL);
+    rconf = ap_get_module_config(r->request_config, &proxy_fcgi_module);
+    if (rconf == NULL) { 
+        rconf = apr_pcalloc(r->pool, sizeof(fcgi_req_config_t));
+        ap_set_module_config(r->request_config, &proxy_fcgi_module, rconf);
+    }
 
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01061)
-                      "set r->path_info to %s", r->path_info);
+    if (NULL != (pathinfo_type = apr_table_get(r->subprocess_env, "proxy-fcgi-pathinfo"))) {
+        /* It has to be on disk for this to work */
+        if (!strcasecmp(pathinfo_type, "full")) { 
+            rconf->need_dirwalk = 1;
+            ap_unescape_url_keep2f(path, 0);
+        }
+        else if (!strcasecmp(pathinfo_type, "first-dot")) { 
+            char *split = ap_strchr(path, '.');
+            if (split) { 
+                char *slash = ap_strchr(split, '/');
+                if (slash) { 
+                    r->path_info = apr_pstrdup(r->pool, slash);
+                    ap_unescape_url_keep2f(r->path_info, 0);
+                    *slash = '\0'; /* truncate path */
+                }
+            }
+        }
+        else if (!strcasecmp(pathinfo_type, "last-dot")) { 
+            char *split = ap_strrchr(path, '.');
+            if (split) { 
+                char *slash = ap_strchr(split, '/');
+                if (slash) { 
+                    r->path_info = apr_pstrdup(r->pool, slash);
+                    ap_unescape_url_keep2f(r->path_info, 0);
+                    *slash = '\0'; /* truncate path */
+                }
+            }
+        }
+        else { 
+            /* before proxy-fcgi-pathinfo had multi-values. This requires the
+             * the FCGI server to fixup PATH_INFO because it's the entire path
+             */
+            r->path_info = apr_pstrcat(r->pool, "/", path, NULL);
+            if (!strcasecmp(pathinfo_type, "unescape")) { 
+                ap_unescape_url_keep2f(r->path_info, 0);
+            }
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01061)
+                    "set r->path_info to %s", r->path_info);
+        }
     }
 
     return OK;
 }
 
-/*
- * Fill in a fastcgi request header with the following type, request id,
- * content length, and padding length.
- *
- * The header array must be at least FCGI_HEADER_LEN bytes long.
- */
-static void fill_in_header(fcgi_header *header,
-                           unsigned char type,
-                           apr_uint16_t request_id,
-                           apr_uint16_t content_len,
-                           unsigned char padding_len)
-{
-    header->version = FCGI_VERSION;
-
-    header->type = type;
-
-    header->requestIdB1 = ((request_id >> 8) & 0xff);
-    header->requestIdB0 = ((request_id) & 0xff);
-
-    header->contentLengthB1 = ((content_len >> 8) & 0xff);
-    header->contentLengthB0 = ((content_len) & 0xff);
-
-    header->paddingLength = padding_len;
-
-    header->reserved = 0;
-}
-
 /* Wrapper for apr_socket_sendv that handles updating the worker stats. */
 static apr_status_t send_data(proxy_conn_rec *conn,
                               struct iovec *vec,
                               int nvec,
-                              apr_size_t *len,
-                              int blocking)
+                              apr_size_t *len)
 {
-    apr_status_t rv = APR_SUCCESS, arv;
+    apr_status_t rv = APR_SUCCESS;
     apr_size_t written = 0, to_write = 0;
     int i, offset;
-    apr_interval_time_t old_timeout;
     apr_socket_t *s = conn->sock;
 
-    if (!blocking) {
-        arv = apr_socket_timeout_get(s, &old_timeout);
-        if (arv != APR_SUCCESS) {
-            return arv;
-        }
-        arv = apr_socket_timeout_set(s, 0);
-        if (arv != APR_SUCCESS) {
-            return arv;
-        }
-    }
-
     for (i = 0; i < nvec; i++) {
         to_write += vec[i].iov_len;
     }
@@ -188,7 +152,7 @@ static apr_status_t send_data(proxy_conn_rec *conn,
     while (to_write) {
         apr_size_t n = 0;
         rv = apr_socket_sendv(s, vec + offset, nvec - offset, &n);
-        if ((rv != APR_SUCCESS) && !APR_STATUS_IS_EAGAIN(rv)) {
+        if (rv != APR_SUCCESS) {
             break;
         }
         if (n > 0) {
@@ -211,12 +175,6 @@ static apr_status_t send_data(proxy_conn_rec *conn,
     conn->worker->s->transferred += written;
     *len = written;
 
-    if (!blocking) {
-        arv = apr_socket_timeout_set(s, old_timeout);
-        if ((arv != APR_SUCCESS) && (rv == APR_SUCCESS)) {
-            return arv;
-        }
-    }
     return rv;
 }
 
@@ -234,58 +192,88 @@ static apr_status_t get_data(proxy_conn_rec *conn,
     return rv;
 }
 
-static apr_status_t send_begin_request(proxy_conn_rec *conn, int request_id)
+static apr_status_t get_data_full(proxy_conn_rec *conn,
+                                  char *buffer,
+                                  apr_size_t buflen)
+{
+    apr_size_t readlen;
+    apr_size_t cumulative_len = 0;
+    apr_status_t rv;
+
+    do {
+        readlen = buflen - cumulative_len;
+        rv = get_data(conn, buffer + cumulative_len, &readlen);
+        if (rv != APR_SUCCESS) {
+            return rv;
+        }
+        cumulative_len += readlen;
+    } while (cumulative_len < buflen);
+
+    return APR_SUCCESS;
+}
+
+static apr_status_t send_begin_request(proxy_conn_rec *conn,
+                                       apr_uint16_t request_id)
 {
     struct iovec vec[2];
-    fcgi_header header;
-    unsigned char farray[FCGI_HEADER_LEN];
-    fcgi_begin_request_body brb;
-    unsigned char abrb[FCGI_HEADER_LEN];
+    ap_fcgi_header header;
+    unsigned char farray[AP_FCGI_HEADER_LEN];
+    ap_fcgi_begin_request_body brb;
+    unsigned char abrb[AP_FCGI_HEADER_LEN];
     apr_size_t len;
 
-    fill_in_header(&header, FCGI_BEGIN_REQUEST, request_id, sizeof(abrb), 0);
+    ap_fcgi_fill_in_header(&header, AP_FCGI_BEGIN_REQUEST, request_id,
+                           sizeof(abrb), 0);
 
-    brb.roleB1 = ((FCGI_RESPONDER >> 8) & 0xff);
-    brb.roleB0 = ((FCGI_RESPONDER) & 0xff);
-    brb.flags = FCGI_KEEP_CONN;
-    brb.reserved[0] = 0;
-    brb.reserved[1] = 0;
-    brb.reserved[2] = 0;
-    brb.reserved[3] = 0;
-    brb.reserved[4] = 0;
+    ap_fcgi_fill_in_request_body(&brb, AP_FCGI_RESPONDER, AP_FCGI_KEEP_CONN);
 
-    fcgi_header_to_array(&header, farray);
-    fcgi_begin_request_body_to_array(&brb, abrb);
+    ap_fcgi_header_to_array(&header, farray);
+    ap_fcgi_begin_request_body_to_array(&brb, abrb);
 
     vec[0].iov_base = (void *)farray;
     vec[0].iov_len = sizeof(farray);
     vec[1].iov_base = (void *)abrb;
     vec[1].iov_len = sizeof(abrb);
 
-    return send_data(conn, vec, 2, &len, 1);
+    return send_data(conn, vec, 2, &len);
 }
 
 static apr_status_t send_environment(proxy_conn_rec *conn, request_rec *r,
-                                     int request_id)
+                                     apr_pool_t *temp_pool,
+                                     apr_uint16_t request_id)
 {
     const apr_array_header_t *envarr;
     const apr_table_entry_t *elts;
     struct iovec vec[2];
-    fcgi_header header;
-    unsigned char farray[FCGI_HEADER_LEN];
-    apr_size_t bodylen, envlen;
-    char *body, *itr;
+    ap_fcgi_header header;
+    unsigned char farray[AP_FCGI_HEADER_LEN];
+    char *body;
     apr_status_t rv;
-    apr_size_t len;
-    int i, numenv;
+    apr_size_t avail_len, len, required_len;
+    int next_elem, starting_elem;
+    char *proxyfilename = r->filename;
+    fcgi_req_config_t *rconf = ap_get_module_config(r->request_config, &proxy_fcgi_module);
+
+    if (rconf) { 
+       if (rconf->need_dirwalk) { 
+          ap_directory_walk(r);
+       }
+    }
+
+    /* Strip balancer prefix */
+    if (r->filename && !strncmp(r->filename, "proxy:balancer://", 17)) { 
+        char *newfname = apr_pstrdup(r->pool, r->filename+17);
+        newfname = ap_strchr(newfname, '/');
+        r->filename = newfname;
+    }
 
     ap_add_common_vars(r);
     ap_add_cgi_vars(r);
+ 
+    r->filename = proxyfilename;
 
     /* XXX are there any FastCGI specific env vars we need to send? */
 
-    bodylen = envlen = 0;
-
     /* XXX mod_cgi/mod_cgid use ap_create_environment here, which fills in
      *     the TZ value specially.  We could use that, but it would mean
      *     parsing the key/value pairs back OUT of the allocated env array,
@@ -293,123 +281,80 @@ static apr_status_t send_environment(proxy_conn_rec *conn, request_rec *r,
      *     place, which would suck. */
 
     envarr = apr_table_elts(r->subprocess_env);
-
     elts = (const apr_table_entry_t *) envarr->elts;
 
-    for (i = 0; i < envarr->nelts; ++i) {
-        apr_size_t keylen, vallen;
-
-        if (! elts[i].key) {
-            continue;
-        }
-
-        keylen = strlen(elts[i].key);
-
-        if (keylen >> 7 == 0) {
-            envlen += 1;
-        }
-        else {
-            envlen += 4;
-        }
-
-        envlen += keylen;
-
-        vallen = strlen(elts[i].val);
-
 #ifdef FCGI_DUMP_ENV_VARS
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01062)
-                      "sending env var '%s' value '%s'",
-                      elts[i].key, elts[i].val);
+    {
+        int i;
+        
+        for (i = 0; i < envarr->nelts; ++i) {
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01062)
+                          "sending env var '%s' value '%s'",
+                          elts[i].key, elts[i].val);
+        }
+    }
 #endif
 
-        if (vallen >> 7 == 0) {
-            envlen += 1;
-        }
-        else {
-            envlen += 4;
-        }
+    /* Send envvars over in as many FastCGI records as it takes, */
+    next_elem = 0; /* starting with the first one */
 
-        envlen += vallen;
+    avail_len = 16 * 1024; /* our limit per record, which could have been up
+                            * to AP_FCGI_MAX_CONTENT_LEN
+                            */
 
-        /* The cast of bodylen is safe since FCGI_MAX_ENV_SIZE is for sure an int */
-        if (envlen > FCGI_MAX_ENV_SIZE) {
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01063)
-                          "truncating environment to %d bytes and %d elements",
-                          (int)bodylen, i);
+    while (next_elem < envarr->nelts) {
+        starting_elem = next_elem;
+        required_len = ap_fcgi_encoded_env_len(r->subprocess_env,
+                                               avail_len,
+                                               &next_elem);
+
+        if (!required_len) {
+            if (next_elem < envarr->nelts) {
+                ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
+                              APLOGNO(02536) "couldn't encode envvar '%s' in %"
+                              APR_SIZE_T_FMT " bytes",
+                              elts[next_elem].key, avail_len);
+                /* skip this envvar and continue */
+                ++next_elem;
+                continue;
+            }
+            /* only an unused element at the end of the array */
             break;
         }
 
-        bodylen = envlen;
+        body = apr_palloc(temp_pool, required_len);
+        rv = ap_fcgi_encode_env(r, r->subprocess_env, body, required_len,
+                                &starting_elem);
+        /* we pre-compute, so we can't run out of space */
+        ap_assert(rv == APR_SUCCESS);
+        /* compute and encode must be in sync */
+        ap_assert(starting_elem == next_elem);
+
+        ap_fcgi_fill_in_header(&header, AP_FCGI_PARAMS, request_id,
+                               (apr_uint16_t)required_len, 0);
+        ap_fcgi_header_to_array(&header, farray);
+
+        vec[0].iov_base = (void *)farray;
+        vec[0].iov_len = sizeof(farray);
+        vec[1].iov_base = body;
+        vec[1].iov_len = required_len;
+
+        rv = send_data(conn, vec, 2, &len);
+        apr_pool_clear(temp_pool);
+
+        if (rv) {
+            return rv;
+        }
     }
 
-    numenv = i;
-
-    body = apr_pcalloc(r->pool, bodylen);
-
-    itr = body;
-
-    for (i = 0; i < numenv; ++i) {
-        apr_size_t keylen, vallen;
-
-        if (! elts[i].key) {
-            continue;
-        }
-
-        keylen = strlen(elts[i].key);
-
-        if (keylen >> 7 == 0) {
-            itr[0] = keylen & 0xff;
-            itr += 1;
-        }
-        else {
-            itr[0] = ((keylen >> 24) & 0xff) | 0x80;
-            itr[1] = ((keylen >> 16) & 0xff);
-            itr[2] = ((keylen >> 8) & 0xff);
-            itr[3] = ((keylen) & 0xff);
-            itr += 4;
-        }
-
-        vallen = strlen(elts[i].val);
-
-        if (vallen >> 7 == 0) {
-            itr[0] = vallen & 0xff;
-            itr += 1;
-        }
-        else {
-            itr[0] = ((vallen >> 24) & 0xff) | 0x80;
-            itr[1] = ((vallen >> 16) & 0xff);
-            itr[2] = ((vallen >> 8) & 0xff);
-            itr[3] = ((vallen) & 0xff);
-            itr += 4;
-        }
-
-        memcpy(itr, elts[i].key, keylen);
-        itr += keylen;
-
-        memcpy(itr, elts[i].val, vallen);
-        itr += vallen;
-    }
-
-    fill_in_header(&header, FCGI_PARAMS, request_id, bodylen, 0);
-    fcgi_header_to_array(&header, farray);
-
-    vec[0].iov_base = (void *)farray;
-    vec[0].iov_len = sizeof(farray);
-    vec[1].iov_base = body;
-    vec[1].iov_len = bodylen;
-
-    rv = send_data(conn, vec, 2, &len, 1);
-    if (rv) {
-        return rv;
-    }
-
-    fill_in_header(&header, FCGI_PARAMS, request_id, 0, 0);
-    fcgi_header_to_array(&header, farray);
+    /* Envvars sent, so say we're done */
+    ap_fcgi_fill_in_header(&header, AP_FCGI_PARAMS, request_id, 0, 0);
+    ap_fcgi_header_to_array(&header, farray);
 
     vec[0].iov_base = (void *)farray;
     vec[0].iov_len = sizeof(farray);
 
-    return send_data(conn, vec, 1, &len, 1);
+    return send_data(conn, vec, 1, &len);
 }
 
 enum {
@@ -427,13 +372,12 @@ enum {
  *
  * Returns 0 if it can't find the end of the headers, and 1 if it found the
  * end of the headers. */
-static int handle_headers(request_rec *r,
-                          int *state,
-                          char *readbuf)
+static int handle_headers(request_rec *r, int *state,
+                          const char *readbuf, apr_size_t readlen)
 {
     const char *itr = readbuf;
 
-    while (*itr) {
+    while (readlen--) {
         if (*itr == '\r') {
             switch (*state) {
                 case HDR_STATE_GOT_CRLF:
@@ -481,83 +425,30 @@ static int handle_headers(request_rec *r,
     return 0;
 }
 
-static void dump_header_to_log(request_rec *r, unsigned char fheader[],
-                               apr_size_t length)
-{
-#ifdef FCGI_DUMP_HEADERS
-    apr_size_t posn = 0;
-    char asc_line[20];
-    char hex_line[60];
-    int i = 0;
-
-    memset(asc_line, 0, sizeof(asc_line));
-    memset(hex_line, 0, sizeof(hex_line));
-
-    while (posn < length) {
-        unsigned char c = fheader[posn];
-
-        if (i >= 20) {
-            i = 0;
-
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01064)
-                          "HEADER: %s %s", asc_line, hex_line);
-
-            memset(asc_line, 0, sizeof(asc_line));
-            memset(hex_line, 0, sizeof(hex_line));
-        }
-
-        if (isprint(c)) {
-            asc_line[i] = c;
-        }
-        else {
-            asc_line[i] = '.';
-        }
-
-        if ((c >> 4) >= 10) {
-            hex_line[i * 3] = 'a' + ((c >> 4) - 10);
-        }
-        else {
-            hex_line[i * 3] = '0' + (c >> 4);
-        }
-
-        if ((c & 0x0F) >= 10) {
-            hex_line[i * 3 + 1] = 'a' + ((c & 0x0F) - 10);
-        }
-        else {
-            hex_line[i * 3 + 1] = '0' + (c & 0xF);
-        }
-
-        hex_line[i * 3 + 2] = ' ';
-
-        i++;
-        posn++;
-    }
-
-    if (i != 1) {
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01065) "HEADER: %s %s",
-                      asc_line, hex_line);
-    }
-
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01066) "HEADER: -EOH-");
-#endif
-}
-
 static apr_status_t dispatch(proxy_conn_rec *conn, proxy_dir_conf *conf,
-                             request_rec *r, int request_id)
+                             request_rec *r, apr_pool_t *setaside_pool,
+                             apr_uint16_t request_id,
+                             const char **err)
 {
     apr_bucket_brigade *ib, *ob;
-    int seen_end_of_headers = 0, done = 0;
+    int seen_end_of_headers = 0, done = 0, ignore_body = 0;
     apr_status_t rv = APR_SUCCESS;
     int script_error_status = HTTP_OK;
     conn_rec *c = r->connection;
     struct iovec vec[2];
-    fcgi_header header;
-    unsigned char farray[FCGI_HEADER_LEN];
+    ap_fcgi_header header;
+    unsigned char farray[AP_FCGI_HEADER_LEN];
     apr_pollfd_t pfd;
     int header_state = HDR_STATE_READING_HEADERS;
-    apr_pool_t *setaside_pool;
+    char stack_iobuf[AP_IOBUFSIZE];
+    apr_size_t iobuf_size = AP_IOBUFSIZE;
+    char *iobuf = stack_iobuf;
 
-    apr_pool_create(&setaside_pool, r->pool);
+    *err = NULL;
+    if (conn->worker->s->io_buffer_size_set) {
+        iobuf_size = conn->worker->s->io_buffer_size;
+        iobuf = apr_palloc(r->pool, iobuf_size);
+    }
 
     pfd.desc_type = APR_POLL_SOCKET;
     pfd.desc.s = conn->sock;
@@ -568,34 +459,33 @@ static apr_status_t dispatch(proxy_conn_rec *conn, proxy_dir_conf *conf,
     ob = apr_brigade_create(r->pool, c->bucket_alloc);
 
     while (! done) {
-        apr_interval_time_t timeout = conn->worker->s->timeout;
+        apr_interval_time_t timeout;
         apr_size_t len;
         int n;
 
         /* We need SOME kind of timeout here, or virtually anything will
          * cause timeout errors. */
-        if (! conn->worker->s->timeout_set) {
-            timeout = apr_time_from_sec(30);
-        }
+        apr_socket_timeout_get(conn->sock, &timeout);
 
         rv = apr_poll(&pfd, 1, &n, timeout);
         if (rv != APR_SUCCESS) {
             if (APR_STATUS_IS_EINTR(rv)) {
                 continue;
             }
+            *err = "polling";
             break;
         }
 
         if (pfd.rtnevents & APR_POLLOUT) {
-            char writebuf[AP_IOBUFSIZE];
-            apr_size_t writebuflen;
+            apr_size_t to_send, writebuflen;
             int last_stdin = 0;
-            int nvec = 0;
+            char *iobuf_cursor;
 
             rv = ap_get_brigade(r->input_filters, ib,
                                 AP_MODE_READBYTES, APR_BLOCK_READ,
-                                sizeof(writebuf));
+                                iobuf_size);
             if (rv != APR_SUCCESS) {
+                *err = "reading input brigade";
                 break;
             }
 
@@ -603,30 +493,48 @@ static apr_status_t dispatch(proxy_conn_rec *conn, proxy_dir_conf *conf,
                 last_stdin = 1;
             }
 
-            writebuflen = sizeof(writebuf);
+            writebuflen = iobuf_size;
 
-            rv = apr_brigade_flatten(ib, writebuf, &writebuflen);
+            rv = apr_brigade_flatten(ib, iobuf, &writebuflen);
 
             apr_brigade_cleanup(ib);
 
             if (rv != APR_SUCCESS) {
+                *err = "flattening brigade";
                 break;
             }
 
-            fill_in_header(&header, FCGI_STDIN, request_id,
-                           (apr_uint16_t) writebuflen, 0);
-            fcgi_header_to_array(&header, farray);
+            to_send = writebuflen;
+            iobuf_cursor = iobuf;
+            while (to_send > 0) {
+                int nvec = 0;
+                apr_size_t write_this_time;
 
-            vec[nvec].iov_base = (void *)farray;
-            vec[nvec].iov_len = sizeof(farray);
-            ++nvec;
-            if (writebuflen) {
-                vec[nvec].iov_base = writebuf;
-                vec[nvec].iov_len = writebuflen;
+                write_this_time =
+                    to_send < AP_FCGI_MAX_CONTENT_LEN ? to_send : AP_FCGI_MAX_CONTENT_LEN;
+
+                ap_fcgi_fill_in_header(&header, AP_FCGI_STDIN, request_id,
+                                       (apr_uint16_t)write_this_time, 0);
+                ap_fcgi_header_to_array(&header, farray);
+
+                vec[nvec].iov_base = (void *)farray;
+                vec[nvec].iov_len = sizeof(farray);
                 ++nvec;
-            }
+                if (writebuflen) {
+                    vec[nvec].iov_base = iobuf_cursor;
+                    vec[nvec].iov_len = write_this_time;
+                    ++nvec;
+                }
 
-            rv = send_data(conn, vec, nvec, &len, 0);
+                rv = send_data(conn, vec, nvec, &len);
+                if (rv != APR_SUCCESS) {
+                    *err = "sending stdin";
+                    break;
+                }
+
+                to_send -= write_this_time;
+                iobuf_cursor += write_this_time;
+            }
             if (rv != APR_SUCCESS) {
                 break;
             }
@@ -634,65 +542,52 @@ static apr_status_t dispatch(proxy_conn_rec *conn, proxy_dir_conf *conf,
             if (last_stdin) {
                 pfd.reqevents = APR_POLLIN; /* Done with input data */
 
-                if (writebuflen) { /* empty FCGI_STDIN not already sent? */
-                    fill_in_header(&header, FCGI_STDIN, request_id, 0, 0);
-                    fcgi_header_to_array(&header, farray);
+                /* signal EOF (empty FCGI_STDIN) */
+                ap_fcgi_fill_in_header(&header, AP_FCGI_STDIN, request_id,
+                                       0, 0);
+                ap_fcgi_header_to_array(&header, farray);
 
-                    vec[0].iov_base = (void *)farray;
-                    vec[0].iov_len = sizeof(farray);
+                vec[0].iov_base = (void *)farray;
+                vec[0].iov_len = sizeof(farray);
 
-                    rv = send_data(conn, vec, 1, &len, 1);
+                rv = send_data(conn, vec, 1, &len);
+                if (rv != APR_SUCCESS) {
+                    *err = "sending empty stdin";
+                    break;
                 }
             }
         }
 
         if (pfd.rtnevents & APR_POLLIN) {
-            /* readbuf has one byte on the end that is always 0, so it's
-             * able to work with a strstr when we search for the end of
-             * the headers, even if we fill the entire length in the recv. */
-            char readbuf[AP_IOBUFSIZE + 1];
             apr_size_t readbuflen;
-            apr_size_t clen;
-            int rid, type;
+            apr_uint16_t clen, rid;
             apr_bucket *b;
-            char plen;
-
-            memset(readbuf, 0, sizeof(readbuf));
-            memset(farray, 0, sizeof(farray));
+            unsigned char plen;
+            unsigned char type, version;
 
             /* First, we grab the header... */
-            readbuflen = FCGI_HEADER_LEN;
-
-            rv = get_data(conn, (char *) farray, &readbuflen);
+            rv = get_data_full(conn, (char *) farray, AP_FCGI_HEADER_LEN);
             if (rv != APR_SUCCESS) {
-                break;
-            }
-
-            dump_header_to_log(r, farray, readbuflen);
-
-            if (readbuflen != FCGI_HEADER_LEN) {
                 ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01067)
-                              "Failed to read entire header "
-                              "got %" APR_SIZE_T_FMT " wanted %d",
-                              readbuflen, FCGI_HEADER_LEN);
-                rv = APR_EINVAL;
+                              "Failed to read FastCGI header");
                 break;
             }
 
-            fcgi_header_from_array(&header, farray);
+#ifdef FCGI_DUMP_HEADERS
+            ap_log_rdata(APLOG_MARK, APLOG_DEBUG, r, "FastCGI header",
+                         farray, AP_FCGI_HEADER_LEN, 0);
+#endif
 
-            if (header.version != FCGI_VERSION) {
+            ap_fcgi_header_fields_from_array(&version, &type, &rid,
+                                             &clen, &plen, farray);
+
+            if (version != AP_FCGI_VERSION_1) {
                 ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01068)
-                              "Got bogus version %d", (int) header.version);
+                              "Got bogus version %d", (int)version);
                 rv = APR_EINVAL;
                 break;
             }
 
-            type = header.type;
-
-            rid = header.requestIdB1 << 8;
-            rid |= header.requestIdB0;
-
             if (rid != request_id) {
                 ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01069)
                               "Got bogus rid %d, expected %d",
@@ -701,14 +596,9 @@ static apr_status_t dispatch(proxy_conn_rec *conn, proxy_dir_conf *conf,
                 break;
             }
 
-            clen = header.contentLengthB1 << 8;
-            clen |= header.contentLengthB0;
-
-            plen = header.paddingLength;
-
 recv_again:
-            if (clen > sizeof(readbuf) - 1) {
-                readbuflen = sizeof(readbuf) - 1;
+            if (clen > iobuf_size) {
+                readbuflen = iobuf_size;
             } else {
                 readbuflen = clen;
             }
@@ -717,24 +607,25 @@ recv_again:
              * recv call, this will eventually change when we move to real
              * nonblocking recv calls. */
             if (readbuflen != 0) {
-                rv = get_data(conn, readbuf, &readbuflen);
+                rv = get_data(conn, iobuf, &readbuflen);
                 if (rv != APR_SUCCESS) {
+                    *err = "reading response body";
                     break;
                 }
-                readbuf[readbuflen] = 0;
             }
 
             switch (type) {
-            case FCGI_STDOUT:
+            case AP_FCGI_STDOUT:
                 if (clen != 0) {
-                    b = apr_bucket_transient_create(readbuf,
+                    b = apr_bucket_transient_create(iobuf,
                                                     readbuflen,
                                                     c->bucket_alloc);
 
                     APR_BRIGADE_INSERT_TAIL(ob, b);
 
                     if (! seen_end_of_headers) {
-                        int st = handle_headers(r, &header_state, readbuf);
+                        int st = handle_headers(r, &header_state,
+                                                iobuf, readbuflen);
 
                         if (st == 1) {
                             int status;
@@ -750,9 +641,16 @@ recv_again:
                                 APR_BRIGADE_INSERT_TAIL(ob, tmp_b);
                                 r->status = status;
                                 ap_pass_brigade(r->output_filters, ob);
-                                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01070)
-                                              "Error parsing script headers");
-                                rv = APR_EINVAL;
+                                if (status == HTTP_NOT_MODIFIED) {
+                                    /* The 304 response MUST NOT contain
+                                     * a message-body, ignore it. */
+                                    ignore_body = 1;
+                                }
+                                else {
+                                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01070)
+                                                    "Error parsing script headers");
+                                    rv = APR_EINVAL;
+                                }
                                 break;
                             }
 
@@ -770,9 +668,14 @@ recv_again:
                                 r->status = HTTP_OK;
                             }
 
-                            if (script_error_status == HTTP_OK) {
+                            if (script_error_status == HTTP_OK
+                                && !APR_BRIGADE_EMPTY(ob) && !ignore_body) {
+                                /* Send the part of the body that we read while
+                                 * reading the headers.
+                                 */
                                 rv = ap_pass_brigade(r->output_filters, ob);
                                 if (rv != APR_SUCCESS) {
+                                    *err = "passing brigade to output filters";
                                     break;
                                 }
                             }
@@ -794,16 +697,17 @@ recv_again:
                          * but that could be a huge amount of data; so we pass
                          * along smaller chunks
                          */
-                        if (script_error_status == HTTP_OK) {
+                        if (script_error_status == HTTP_OK && !ignore_body) {
                             rv = ap_pass_brigade(r->output_filters, ob);
                             if (rv != APR_SUCCESS) {
+                                *err = "passing brigade to output filters";
                                 break;
                             }
                         }
                         apr_brigade_cleanup(ob);
                     }
 
-                    /* If we didn't read all the data go back and get the
+                    /* If we didn't read all the data, go back and get the
                      * rest of it. */
                     if (clen > readbuflen) {
                         clen -= readbuflen;
@@ -817,6 +721,7 @@ recv_again:
                         APR_BRIGADE_INSERT_TAIL(ob, b);
                         rv = ap_pass_brigade(r->output_filters, ob);
                         if (rv != APR_SUCCESS) {
+                            *err = "passing brigade to output filters";
                             break;
                         }
                     }
@@ -825,11 +730,11 @@ recv_again:
                 }
                 break;
 
-            case FCGI_STDERR:
+            case AP_FCGI_STDERR:
                 /* TODO: Should probably clean up this logging a bit... */
                 if (clen) {
                     ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(01071)
-                                  "Got error '%s'", readbuf);
+                                  "Got error '%.*s'", (int)readbuflen, iobuf);
                 }
 
                 if (clen > readbuflen) {
@@ -838,7 +743,7 @@ recv_again:
                 }
                 break;
 
-            case FCGI_END_REQUEST:
+            case AP_FCGI_END_REQUEST:
                 done = 1;
                 break;
 
@@ -847,12 +752,16 @@ recv_again:
                               "Got bogus record %d", type);
                 break;
             }
+            /* Leave on above switch's inner error. */
+            if (rv != APR_SUCCESS) {
+                break;
+            }
 
             if (plen) {
-                readbuflen = plen;
-
-                rv = get_data(conn, readbuf, &readbuflen);
+                rv = get_data_full(conn, iobuf, plen);
                 if (rv != APR_SUCCESS) {
+                    ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(02537)
+                                  "Error occurred reading padding");
                     break;
                 }
             }
@@ -880,14 +789,16 @@ static int fcgi_do_request(apr_pool_t *p, request_rec *r,
                            char *url, char *server_portstr)
 {
     /* Request IDs are arbitrary numbers that we assign to a
-     * single request. This would allow multiplex/pipelinig of
+     * single request. This would allow multiplex/pipelining of
      * multiple requests to the same FastCGI connection, but
      * we don't support that, and always use a value of '1' to
      * keep things simple. */
-    int request_id = 1;
+    apr_uint16_t request_id = 1;
     apr_status_t rv;
+    apr_pool_t *temp_pool;
+    const char *err;
 
-    /* Step 1: Send FCGI_BEGIN_REQUEST */
+    /* Step 1: Send AP_FCGI_BEGIN_REQUEST */
     rv = send_begin_request(conn, request_id);
     if (rv != APR_SUCCESS) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(01073)
@@ -896,8 +807,10 @@ static int fcgi_do_request(apr_pool_t *p, request_rec *r,
         return HTTP_SERVICE_UNAVAILABLE;
     }
 
+    apr_pool_create(&temp_pool, r->pool);
+
     /* Step 2: Send Environment via FCGI_PARAMS */
-    rv = send_environment(conn, r, request_id);
+    rv = send_environment(conn, r, temp_pool, request_id);
     if (rv != APR_SUCCESS) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(01074)
                       "Failed writing Environment to %s:", server_portstr);
@@ -906,10 +819,14 @@ static int fcgi_do_request(apr_pool_t *p, request_rec *r,
     }
 
     /* Step 3: Read records from the back end server and handle them. */
-    rv = dispatch(conn, conf, r, request_id);
+    rv = dispatch(conn, conf, r, temp_pool, request_id, &err);
     if (rv != APR_SUCCESS) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(01075)
-                      "Error dispatching request to %s:", server_portstr);
+                      "Error dispatching request to %s: %s%s%s",
+                      server_portstr,
+                      err ? "(" : "",
+                      err ? err : "",
+                      err ? ")" : "");
         conn->close = 1;
         return HTTP_SERVICE_UNAVAILABLE;
     }
@@ -943,10 +860,7 @@ static int proxy_fcgi_handler(request_rec *r, proxy_worker *worker,
                   "url: %s proxyname: %s proxyport: %d",
                  url, proxyname, proxyport);
 
-    if (strncasecmp(url, "fcgi:", 5) == 0) {
-        url += 5;
-    }
-    else {
+    if (strncasecmp(url, "fcgi:", 5) != 0) {
         ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01077) "declining URL %s", url);
         return DECLINED;
     }
@@ -954,16 +868,14 @@ static int proxy_fcgi_handler(request_rec *r, proxy_worker *worker,
     ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(01078) "serving URL %s", url);
 
     /* Create space for state information */
-    if (! backend) {
-        status = ap_proxy_acquire_connection(FCGI_SCHEME, &backend, worker,
-                                             r->server);
-        if (status != OK) {
-            if (backend) {
-                backend->close = 1;
-                ap_proxy_release_connection(FCGI_SCHEME, backend, r->server);
-            }
-            return status;
+    status = ap_proxy_acquire_connection(FCGI_SCHEME, &backend, worker,
+                                         r->server);
+    if (status != OK) {
+        if (backend) {
+            backend->close = 1;
+            ap_proxy_release_connection(FCGI_SCHEME, backend, r->server);
         }
+        return status;
     }
 
     backend->is_ssl = 0;
@@ -977,11 +889,15 @@ static int proxy_fcgi_handler(request_rec *r, proxy_worker *worker,
         goto cleanup;
     }
 
-    /* XXX Setting close to 0 is a great way to end up with
-     *     timeouts at this point, since we lack good ways to manage the
-     *     back end fastcgi processes.  This should be revisited when we
-     *     have a better story on that part of things. */
+    /* This scheme handler does not reuse connections by default, to
+     * avoid tieing up a fastcgi that isn't expecting to work on 
+     * parallel requests.  But if the user went out of their way to
+     * type the default value of disablereuse=off, we'll allow it.
+     */  
     backend->close = 1;
+    if (worker->s->disablereuse_set && !worker->s->disablereuse) { 
+        backend->close = 0;
+    }
 
     /* Step Two: Make the Connection */
     if (ap_proxy_connect_backend(FCGI_SCHEME, backend, worker, r->server)) {
